# FILE: coords.tcl
#
# Routines to draw coordinate axes.
#
# These routines could be wrapped up in an Oc_Class, but were originally
# written before the existence of Oc or em classes.
#
# Rendering of text is can be slow if a font server across a network
# is accessed, so rather than redraw the coordinate axes on each rotation,
# we draw four sets of axis at initialization, stored in different
# corners of the canvas, which is "scrollregion"'ed sized to 3x3 times
# the canvas width/height.  Then on receipt of a show coords request,
# we just scroll the canvas to display the appropriate axes.
#
# The coordinates axes are positioned in the canvas as a whole, as
# illustrated below (I=0, II=90, III=180, IV=270):
#
#        --------------------
#	 |     |     |
#	 |  II |     | 	 I
#        |     |     |
#	 |-------------------
#	 |     |     |
#	 |     |     |
#        |     |     |
#	 |-------------------
#	 |     |     |
#	 | III |     | 	IV
#	 |     |     |
#
#
# NOTE: This routine gets the coords display size from the canvas
#   -height value.  This *must* be set before calling this routine.
#
#   Each element in the coordinate canvas is tagged with one of the 6
# identifiers xp, xm, yp, ym, zp, zm, for "plus x-axis," "minus x-axis",
# etc.  Theses tagges are used to color the coordinates inside the
# coloring proc SetCoordColors.
proc InitCoords { canvas {curangle 0}} {
    $canvas delete all  ;# In case this isn't the first call

    # Check for square canvas viewport
    set size [$canvas cget -height]
    set width [$canvas cget -width]
    if { $size<1 } {
	set size $width
	$canvas configure -height $size
    } elseif {$width != $size} { 
	$canvas configure -width $size
    }

    # Set up size of whole canvas
    set scrollx [expr {3*$size}]
    set scrolly [expr {3*$size}]
    $canvas configure \
	    -scrollregion "0 0 $scrollx $scrolly" \
	    -xscrollincrement $size -yscrollincrement $size \
	    -confine 1

    # Draw each axis configuration.  For better user-apparent
    # responsiveness, draw & show the current angle now, and
    # put the other drawing requests on the event stack.
    for { set rot 0 } { $rot < 4 } { incr rot } {
	switch -- $rot {
	    0 { set xc [expr {2.5*$size}] ; set yc [expr {0.5*$size}] }
	    1 { set xc [expr {0.5*$size}] ; set yc [expr {0.5*$size}] }
	    2 { set xc [expr {0.5*$size}] ; set yc [expr {2.5*$size}] }
	    3 { set xc [expr {2.5*$size}] ; set yc [expr {2.5*$size}] }
	}
	set angle [expr {90*$rot}]
	if { $angle == $curangle } {
	    DrawCoords $canvas $size $xc $yc $angle
	    ShowCoords $canvas $curangle
	} else {
	    after idle "DrawCoords $canvas $size $xc $yc $angle"
	}
    }
}

# Translate canvas in viewport to show corresponding coordinate axes.
proc ShowCoords { canvas angle } {
    set rot [expr {round($angle/90.)%4}]
    switch -- $rot {
        0 { set xfrac 0.6667 ; set yfrac 0.     }
        1 { set xfrac 0.     ; set yfrac 0.     }
        2 { set xfrac 0.     ; set yfrac 0.6667 }
        3 { set xfrac 0.6667 ; set yfrac 0.6667 }
    }
    $canvas xview moveto $xfrac
    $canvas yview moveto $yfrac
}

#
# Draws a 3D coordinate axes with orientation 'angle' (should be one of
# 0, 90, 180 or 270) on 'canvas', inside a square of edge length 'size',
# centered at (xc,yc).  Coordinate colors should be set by a separate
# call to SetCoordColors.
#
# Note 1: It is the responsibility of the calling routine to erase
#         'canvas' *before* calling this routine.
#
# Note 2: These routines have been optimized for size==50.
#
label .temporary_label
set _mmdisp(label_font) [.temporary_label cget -font]
destroy .temporary_label
proc DrawCoords { canvas size xc yc angle } {
    set rot [expr {round($angle/90.)%4}]
    switch -- $rot {
        0 { DrawCoords0   $canvas $size $xc $yc }
        1 { DrawCoords90  $canvas $size $xc $yc }
        2 { DrawCoords180 $canvas $size $xc $yc }
        3 { DrawCoords270 $canvas $size $xc $yc }
    }
    $canvas itemconfig text -font [Oc_Font Get compressed_bold]
}

proc DrawCoordsSetVars { size arrbasename arrsizename deltaname} {
    # Set variables used in DrawCoords* routines
    upvar $arrbasename arrbase
    upvar $arrsizename arrsize
    upvar $deltaname   delta

    set arrbase 5
    set arrsize "$arrbase $arrbase [expr {round($arrbase/2)}]"
    set delta [expr {floor(($size-$arrbase)/2.)}]
}

proc DrawZAxis { canvas size xc yc delta arrbase arrsize } {
    # +Z axis
    set diag [expr {round($delta/1.414)}]
    set x1 [expr {$xc-$diag}];  set y1 [expr {$yc+$diag}]
    $canvas create line $xc $yc $x1 $y1 \
        -arrow last -arrowshape $arrsize -tags zp
    $canvas create text [expr {$xc-$size/2.+2}] [expr {$y1-4}] -anchor nw \
            -text "+z" -tags [list text zp]
    # -Z axis
    set x1 [expr {$xc+$diag}];  set y1 [expr {$yc-$diag}]
    set step [expr {round((($delta-$arrbase)/1.414)/4.)}]
    set xi [expr {$xc}];  set yi [expr {$yc}]
    set xj [expr {$xi+$step+1}];  set yj [expr {$yi-$step-1}]
    $canvas create line $xi $yi $xj $yj -arrow none -tags zm
    set xi [expr {$xj+round($step/2.)}]
    set yi [expr {$yj-round($step/2.)}]
    set xj [expr {$xi+$step}];  set yj [expr {$yi-$step}]
    $canvas create line $xi $yi $xj $yj -arrow none -tags zm
    set xi [expr {$xj+round($step/2.)}]
    set yi [expr {$yj-round($step/2.)}]
    set xj $x1;  set yj $y1
    $canvas create line $xi $yi $xj $yj \
            -arrow last -arrowshape $arrsize -tags zm
    $canvas create text [expr {$xc+$size/2.-1}] [expr {$yj+4}] \
	    -anchor se -text "-z" -tags [list text zm]
}

proc DrawCoords0 { canvas size xc yc } {
    DrawCoordsSetVars $size arrbase arrsize delta
    # Z-axis
    DrawZAxis $canvas $size $xc $yc $delta $arrbase $arrsize
    # X-axis
    set x1 [expr {$xc+$delta}];  set y1 $yc
    $canvas create line $xc $yc $x1 $y1 \
            -arrow last -arrowshape $arrsize -tags xp
    $canvas create text $x1 [expr {$y1+round($arrbase/4.)-1}] -anchor ne \
            -text "+x" -tags [list text xp]
    set x1 [expr {$xc-$delta}]
    $canvas create line $xc $yc $x1 $y1 \
            -arrow none -arrowshape $arrsize -tags xm
    # Y-axis
    set x1 $xc;  set y1 [expr {$yc-$delta}];
    $canvas create line $xc $yc $x1 $y1 \
            -arrow last -arrowshape $arrsize -tags yp
    $canvas create text [expr {$x1-round($arrbase/2.)}] $y1 -anchor ne \
            -text "+y" -tags [list text yp]
    set y1 [expr {$yc+$delta}];
    $canvas create line $xc $yc $x1 $y1 \
            -arrow none -arrowshape $arrsize -tags ym
}

proc DrawCoords90 { canvas size xc yc } {
    DrawCoordsSetVars $size arrbase arrsize delta
    # Z-axis
    DrawZAxis $canvas $size $xc $yc $delta $arrbase $arrsize
    # X-axis
    set x1 $xc;  set y1 [expr {$yc-$delta}]
    $canvas create line $xc $yc $x1 $y1 \
            -arrow last -arrowshape $arrsize -tags xp
    $canvas create text [expr {$x1-round($arrbase/2.)}] \
	    [expr {$yc-$size/2.-2}] \
	    -anchor ne -text "+x" -tags [list text xp]
    set y1 [expr {$yc+$delta}]
    $canvas create line $xc $yc $x1 $y1 \
            -arrow none -arrowshape $arrsize -tags xm
    # Y-axis
    set x1 [expr {$xc-$delta}];  set y1 $yc
    $canvas create line $xc $yc $x1 $y1 \
            -arrow last -arrowshape $arrsize -tags yp
    $canvas create text $x1 [expr {$y1-round($arrbase/4.)}] \
	    -anchor sw -text "+y" -tags [list text yp]
    set x1 [expr {$xc+$delta}]
    $canvas create line $xc $yc $x1 $y1 \
            -arrow none -arrowshape $arrsize -tags ym
}


proc DrawCoords180 { canvas size xc yc } {
    DrawCoordsSetVars $size arrbase arrsize delta
    # Z-axis
    DrawZAxis $canvas $size $xc $yc $delta $arrbase $arrsize
    # X-axis
    set x1 [expr {$xc-$delta}];  set y1 $yc
    $canvas create line $xc $yc $x1 $y1 \
            -arrow last -arrowshape $arrsize -tags xp
    $canvas create text $x1 [expr {$y1-round($arrbase/4.)}] -anchor sw \
            -text "+x" -tags [list text xp]
    set x1 [expr {$xc+$delta}]
    $canvas create line $xc $yc $x1 $y1 \
            -arrow none -arrowshape $arrsize -tags xm
    # Y-axis
    set x1 $xc;  set y1 [expr {$yc+$delta}]
    $canvas create line $xc $yc $x1 $y1 \
            -arrow last -arrowshape $arrsize -tags yp
    $canvas create text [expr {$x1+$arrbase}] [expr {$yc+$size/2.-1}] \
	    -anchor sw -text "+y" -tags [list text yp]
    set y1 [expr {$yc-$delta}]
    $canvas create line $xc $yc $x1 $y1 \
            -arrow none -arrowshape $arrsize -tags ym
}

proc DrawCoords270 { canvas size xc yc } {
    DrawCoordsSetVars $size arrbase arrsize delta
    # Z-axis
    DrawZAxis $canvas $size $xc $yc $delta $arrbase $arrsize
    # X-axis
    set x1 $xc;  set y1 [expr {$yc+$delta}];
    $canvas create line $xc $yc $x1 $y1 \
            -arrow last -arrowshape $arrsize -tags xp
    $canvas create text [expr {$x1+$arrbase-1}] [expr {$yc+$size/2.+1}] \
	    -anchor sw -text "+x" -tags [list text xp]
    set y1 [expr {$yc-$delta}];
    $canvas create line $xc $yc $x1 $y1 \
            -arrow none -arrowshape $arrsize -tags xm
    # Y-axis
    set x1 [expr {$xc+$delta}];  set y1 $yc
    $canvas create line $xc $yc $x1 $y1 \
            -arrow last -arrowshape $arrsize -tags yp
    $canvas create text [expr {$xc+$size/2.-1}] \
	    [expr {$y1+round($arrbase/4.)-2}] \
	    -anchor ne -text "+y" -tags [list text yp]
    set x1 [expr {$xc-$delta}]
    $canvas create line $xc $yc $x1 $y1 \
            -arrow none -arrowshape $arrsize -tags ym
}

proc SetCoordColors { canvas {bgcolor {}} {axiscolors {}}} {
    # axiscolors should be a 6 element list of colors:
    #   +x color, -x color, +y color, -y color, +z color, -z color
    # Any empty or missing color will be painted black, unless bgcolor
    # is black in which case the default color will be white.

    if {[string match {} $bgcolor]} { set bgcolor white }
    $canvas configure -bg $bgcolor

    set defaultcolor black
    if {[string match black $bgcolor]} { set defaultcolor white }
    foreach { xp xm yp ym zp zm } $axiscolors {}
    foreach tag { xp xm yp ym zp zm } {
        if {![info exists $tag] || \
                [string match {} [set color [set $tag]]]} {
            set color $defaultcolor
        }
        $canvas itemconfig $tag -fill $color
    }
}
